#
# ImageViewCanvasTypesGtk.py -- drawing classes for ImageViewCanvas widget
# 
# Eric Jeschke (eric@naoj.org)
#
# Copyright (c) Eric R. Jeschke.  All rights reserved.
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

# All gtk drawing in Ginga is done with cairo
from ginga.cairow.ImageViewCanvasTypesCairo import *


#END
