#
# convenience module for importing canvas mixin classes
#
from ginga.canvas.CanvasObject import *
from ginga.canvas.CompoundMixin import CompoundMixin
from ginga.canvas.CanvasMixin import CanvasMixin
from ginga.canvas.DrawingMixin import DrawingMixin

#END
